"""
General Layout shortcuts.

Layouts are decorators that receive a parent ElementTOSC, ControlType 
for children, size and color and create a bunch of children whose 
frames align inside the parent's frame.

To define a layout, add one of the preset layouts: layoutColumn, 
layoutRow, etc.to your function and set that function to receive a 
list of ElementTOSC. Then return a list[Property] (Properties)
that you want to add to the layout.

The order is:

1. Calculate Frame and Color arrays
2. Create the children Controls and transform them.
3. Then the decorated function happens.
4. You can return a list[Property] (Properties) to apply to the layout.
5. Then the layout is returned to whoever called the decorated function.

Because your function will be executed towards the end, you can add
any number of nested layouts to the children as they will have their 
frame and color set already.

The decorated function is also a good place to add Values and Messages
to the children.

All Layouts are currently built in top > bottom, left > right order.
"""

from copy import deepcopy
from typing import Any, Callable
from .tosc import ElementTOSC
from .elements import (
    ControlElements,
    controlType,
    ControlType,
)
from .controls import (
    Properties,
)
import numpy as np


"""

ARRANGE AND LAYOUT

"""

def colorChecker(color: Any) -> tuple:
    """Allow for passing rgba in 0-255, 0-1 and hex, then convert to 0-1"""
    if isinstance(color[0], int):
        return tuple(i / 255 for i in color)
    elif isinstance(color[0], float):
        return color
    elif isinstance(color[0], str):
        color = color.replace("#", "")
        return (
            tuple(int(color[i : i + 2], 16) / 255 for i in (0, 2, 4, 6))
            if len(color) > 7
            else tuple(int(color[i : i + 2], 16) / 255 for i in (0, 2, 4)) + (1,)
        )
    else:
        raise TypeError(f"{color} type is not a valid color.")


"""

LAYOUT FUNCTIONS

"""


def Layout(
    layout: ElementTOSC,
    controlT: controlType,
    F: np.ndarray,
    C: np.ndarray,
    func: Callable[[list[ElementTOSC]], Properties],
):
    """Basic process to append multiple properties to a layout of controls"""

    children = [ElementTOSC(layout.createChild(controlT)) for i in range(F.shape[0])]
    for g, f, c in zip(children, F, C):
        g.setFrame(f.astype(int))
        g.setColor(c)

    # Add extra properties to the parent, optional return
    properties: Properties = func(children)
    if properties is not None:
        [layout.createProperty(p) for p in properties]

    return layout


def column(func):
    """Create a column of groups with a color gradient.

    Args:
        parent: The Control that becomes the layout.
        controlType: The Control type of the generated children.
        size: The size and ratio of the columns, (1,2,1) means 3 columns with 1:2:1 ratio.
        colors: Tuple of two colors for linear gradient.
    """

    def wrapper(
        parent: ElementTOSC,
        controlType: controlType,
        size: tuple = (1, 2, 1),
        colors: tuple = (
            (0.25, 0.25, 0.25, 1.0), (0.25, 0.25, 0.25, 1.0)),
    ):
        colors = tuple(colorChecker(i) for i in colors)  # makes sure is normalized
        frame = parent.getFrame()

        H = frame[3] * (np.asarray(size) / np.sum(size))
        W = [frame[2] for i in size]
        Y = np.cumsum(np.concatenate(([0], H)))[:-1]
        X = np.resize((0), len(size))
        F = np.asarray((X, Y, W, H, X)).T
        C = np.linspace(colors[0], colors[1], len(size))
        return Layout(parent, controlType, F, C, func)

    return wrapper


def row(func):
    """Create a row of groups with a color gradient.

    Args:
        parent: The Control that becomes the layout.
        controlType: The Control type of the generated children.
        size: The size and ratio of the rows, (1,2,1) means 3 rows with 1:2:1 ratio.
        colors: Tuple of two colors for linear gradient.
    """

    def wrapper(
        parent: ElementTOSC,
        controlType: ControlType,
        size: tuple = (1, 2, 1),
        frame: tuple = (0, 0, 1600, 640),
        colors: tuple | str = ((0.25, 0.25, 0.25, 1.0), (0.25, 0.25, 0.25, 1.0)),
    ):
        colors = tuple(colorChecker(i) for i in colors)  # makes sure is normalized
        frame = parent.getFrame()

        H = np.resize(frame[3], len(size)) # type: ignore
        W = frame[2] * (np.asarray(size) / np.sum(size))
        Y = np.resize((0), len(size))
        X = np.cumsum(np.concatenate(([0], W)))[:-1]
        F = np.asarray((X, Y, W, H, X)).T
        C = np.linspace(colors[0], colors[1], len(size))
        return Layout(parent, controlType, F, C, func)

    return wrapper


def grid(func):
    """Create an x*y grid of groups.

    Args:
        parent: The Control that becomes the layout.
        controlType: The Control type of the generated children.
        size: the row x column size in tuple, ej (4,4) or (5, 3), etc
        colors: tuple of two colors gradient, see colorStyle
        colorStyle:
            Select a gradient style.
            0: horizontal gradient
            1: vetical gradient
            2: sequential gradient 1
            3: sequential gradient 2
            >= 4: centered/mirrored gradient, moves position with number

    """

    def wrapper(
        parent: ElementTOSC,
        controlType: ControlType,
        size: tuple = (4, 4),
        colors: tuple=  (
            (0.25, 0.25, 0.25, 1.0),
            (0.5, 0.5, 0.5, 1.0),
        ),
        colorStyle: int = 0,
    ):

        if (frame:=parent.getFrame()) is None:
            raise ValueError(f"{parent} has no frame.")

        colors = tuple(colorChecker(i) for i in colors)

        w = frame[2] / size[0]
        h = frame[3] / size[1]
        M = np.asarray(
            tuple(
                (row, column)
                for row in np.arange(stop=frame[2], step=w) # type: ignore
                for column in np.arange(stop=frame[3], step=h) # type: ignore
            )
        ).T

        X = M[0]
        Y = M[1]
        W = np.repeat(w, X.size)
        H = np.repeat(h, Y.size)
        F = np.asarray((X, Y, W, H, X)).T

        # TO DO : Optimize
        if colorStyle == 0:  # horizontal
            C = np.linspace(colors[0], colors[1], size[0])
            C = np.repeat(C.T, size[1]).T.reshape(4, size[0] * size[1]).T
        elif colorStyle == 1:  # vertical
            C = np.linspace(colors[0], colors[1], size[1])
            C = np.asarray(np.resize(C, size[0] * C.size)).reshape(size[0] * size[1], 4)
        elif colorStyle == 2:  # sequential
            C = np.linspace(colors[0], colors[1], size[0] * size[1])
        elif colorStyle == 3:  # sequential inverted
            C = np.linspace(colors[0], colors[1], size[0] * size[1])
            C = np.asarray([C[i :: size[0]] for i in range(size[0])]).reshape(
                size[0] * size[1], 4
            )
        elif colorStyle >= 4:  # centered / mirrored
            C = np.linspace(colors[0], colors[1], size[0] * size[1])
            C = np.roll(C, colorStyle * 4)
            C[0:colorStyle] = np.flip(
                np.roll(C, (-1 - colorStyle) * 4)[0:colorStyle], axis=0
            )
        else:
            raise ValueError(f"{colorStyle} is not a valid colorStyle.")

        return Layout(parent, controlType, F, C, func)

    return wrapper



"""
API for TOSC Control Elements.
"""
import logging
from copy import deepcopy
import sys
import re
from typing import Callable, Generic, NewType, TypeAlias, TypeGuard, TypeVar
import zlib
import uuid
from typing import Any

from pyparsing import Optional
from .elements import (
    Partial,
    Trigger,
    Property,
    Value,
    OSC,
    MIDI,
    LOCAL,
    MidiMessage,
    PropertyType,
    ControlElements,
    ControlType,
    PropertyFactory,
    elementType
)
from .controls import (
    ControlConverter,
    ControlFactory,
    Message,
    XmlFactory,
    Properties,
    controlType,
    Control,
)

import xml.etree.ElementTree as ET
# from lxml import etree as ET

ElementXML: TypeAlias = ET.Element


def simpleProperty(func):
    """Pass value as text arg"""

    def wrapper(self: "ElementTOSC", value):
        type, key = func(self)
        element = self.getProperty(key)
        if element is not None:
            self.properties.remove(element)
        return self.createPropertyUnsafe(
            Property(type.value, key, str(value)))

    return wrapper

def booleanProperty(func):
    """Pass value as bool, so outline is True"""

    def wrapper(self: "ElementTOSC", value):
        type, key = func(self)
        element = self.getProperty(key)
        if element is not None:
            self.properties.remove(element)
        return self.createPropertyUnsafe(
            Property(type.value, key, repr(int(value)))
        )

    return wrapper

def multiProperty(func):
    """Pass args as tuple of keys, so color is ("r","g","b","a")"""

    def wrapper(self: "ElementTOSC", params):
        type, key, paramKeys = func(self)
        element = self.getProperty(key)
        if element is not None:
            self.properties.remove(element)
        return self.createPropertyUnsafe(
            Property(
                type.value,
                key,
                params={k: repr(params[i])
                        for i, k in enumerate(paramKeys)},
            )
        )

    return wrapper


class ElementTOSC:
    """
    Contains a Node Element and its SubElements. Creates them if not found.
    """

    def __init__(self, e: ElementXML):
        """Find SubElements on init

        Args:
            e (ET.Element): <node> Element

        Attributes:
            properties (ET.Element): Find <properties>
            values (ET.Element): Find <values>
            messages (ET.Element): Find <messages>
            children (ET.Element): Find <children>
        """
        self.node = e
        self.properties = self.getCreate("properties")
        self.values = self.getCreate("values")
        self.messages = self.getCreate("messages")
        self.children = self.getCreate("children")

    def __iter__(self):
        """Return iter over children"""
        return iter(self.children)

    def __getitem__(self,item):
        return self.__class__(self.children[item])

    def append(self, e: "ElementTOSC") -> "ElementTOSC":
        """Append an ElementTOSC's Node to this element's Children"""
        self.children.append(e.node)
        return self

    def getCreate(self, target):
        s = self.node.find(target)
        if s is not None:
            return s
        return ET.SubElement(self.node, target)

    @classmethod
    def fromFile(cls, file: str) -> "ElementTOSC":
        """Load a .tosc file into an XML Element and then as ElementTOSC"""
        return cls(load(file)[0])

    def getProperty(self, key: str) -> ElementXML | None:
        """Finds property that has a key child that matches

        Args:
            key (str): Key text

        Returns:
            ElementXML | None: ./property
        """
        return findKey(self.properties, key)

    def getPropertyValue(self, key: str) -> ElementXML | None:
        """Finds value child of property with given key.

        Args:
            key (str): Key text

        Returns:
            ElementXML | None: ./property/value
        """
        return findKey(self.properties, key).find("value")

    def getPropertyParam(self, key: str, param: str) -> ElementXML | None:
        """Finds value child of property with given key, then
        finds the param child of value.

        Args:
            key (str): Key text
            param (str): Param tag

        Returns:
            ElementXML | None: ./property/value/param
        """
        if (value := findKey(self.properties, key).find("value")) is not None:
            return value.find(param)
        return None

    def hasProperty(self, key: str) -> bool:
        return True if findKey(self.properties, key) else False

    def setProperty(self, key: str, value: str = "",
                    params: dict = {}) -> bool:
        if (e:=findKey(self.properties, key)) is None:
            raise ValueError(f"{key} doesn't exist.")
        return XmlFactory.modifyProperty(e, value, params)

    def createProperty(self, property: Property) -> bool:
        if findKey(self.properties, property.key) is not None:
            raise ValueError(f"{property.key} already exists.")
        return self.createPropertyUnsafe(property)

    def createPropertyUnsafe(self, property: Property) -> bool:
        return XmlFactory.buildProperties(self.properties,[property])

    def getValue(self, key: str) -> ElementXML | None:
        return findKey(self.values, key)

    def getValueParam(self, key: str, param: str) -> ElementXML | None:
        try:
            return findKey(self.values, key).find(param)
        except:
            raise ValueError

    def hasValue(self, key: str) -> bool:
        return True if findKey(self.values, key) is not None else False

    def createValue(self, value: Value) -> bool:
        if findKey(self.values, value.key) is not None:
            raise ValueError(f"{value.key} already exists.")
        return XmlFactory.buildValues(self.values, [value])

    def setValue(self, value: Value) -> bool:
        e = findKey(self.values, value.key)
        if e is None:
            raise ValueError(f"{value.key} doesn't exist.")
        return XmlFactory.modifyValue(e, value)

    def createOSC(self, message: OSC = OSC()) -> bool:
        """Builds and appends an OSC message"""
        # return ControlElements.OSC, message
        return XmlFactory.buildMessages(self.messages, [message])

    def createMIDI(self, message: MIDI = MIDI()) -> bool:
        """Builds and appends a M1D1 message"""
        return XmlFactory.buildMessages(self.messages, [message])

    def createLOCAL(self, message: LOCAL = LOCAL()) -> bool:
        """Builds and appends a LOCAL message"""
        return XmlFactory.buildMessages(self.messages, [message])

    def removeOSC(self) -> bool:
        """Find and remove all OSC messages"""
        [msg.remove for msg in self.messages.findall(ControlElements.OSC.value)]
        return True

    def removeMIDI(self) -> bool:
        """Find and remove all MIDI messages"""
        [msg.remove for msg in self.messages.findall(ControlElements.MIDI.value)]
        return True

    def removeLOCAL(self) -> bool:
        """Find and remove all LOCAL messages"""
        [msg.remove for msg in self.messages.findall(ControlElements.LOCAL.value)]
        return True

    def findChildByName(self, name: str) -> ElementXML | None:
        for child in self.children:
            if child.find(ControlElements.PROPERTIES.value) is None:
                continue
            if re.fullmatch(
                getTextValueFromKey(
                    child.find(ControlElements.PROPERTIES.value), "name"
                ),
                name,
            ):
                return child
        return None

    def createChild(self, type: controlType) -> ElementXML:
        return XmlFactory.buildNode(self.children, type)

    def getID(self) -> str:
        return str(self.node.attrib["ID"])

    def isControlType(self, control: ControlType):
        return str(self.node.attrib["type"]) == control.value

    def setControlType(self, value: ControlType):
        """See ControlType Element"""
        self.node.attrib["type"] = value.value
        return True

    def getFrame(self) -> tuple[int, ...]:
        """Wrapper for getX, getY, etc."""
        return (self.getX(), self.getY(), self.getW(), self.getH(),)

    def getColor(self) -> tuple[float, ...]:
        return (self.getR(), self.getG(), self.getB(), self.getA(),)

    def getR(self) -> float:
        if (p:=self.getPropertyParam("color", "r")) is not None:
            return float(str(p.text))
        raise ValueError("Could not find color: r")

    def getG(self) -> float:
        if (p:=self.getPropertyParam("color", "g")) is not None:
            return float(str(p.text))
        raise ValueError("Could not find color: g")

    def getB(self) -> float:
        if (p:=self.getPropertyParam("color", "b")) is not None:
            return float(str(p.text))
        raise ValueError("Could not find color: b")
        
    def getA(self) -> float:
        if (p:=self.getPropertyParam("color", "a")) is not None:
            return float(str(p.text))
        raise ValueError("Could not find color: a")

    def getX(self) -> int:
        if (p:=self.getPropertyParam("frame", "x")) is not None:
            return int(str(p.text))
        raise ValueError("Could not find frame: x")

    def getY(self) -> int:
        if (p:=self.getPropertyParam("frame", "y")) is not None:
            return int(str(p.text))
        raise ValueError("Could not find frame: y")

    def getW(self) -> int:
        if (p:=self.getPropertyParam("frame", "w")) is not None:
            return int(str(p.text))
        raise ValueError("Could not find frame: w")

    def getH(self) -> int:
        if (p:=self.getPropertyParam("frame", "h")) is not None:
            return int(str(p.text))
        raise ValueError("Could not find frame: h")

    @simpleProperty
    def setName(self):
        """String"""
        return PropertyType.STRING, "name"

    def getName(self):
        return self.getPropertyValue("name").text

    @simpleProperty
    def setTag(self):
        """String"""
        return PropertyType.STRING, "tag"

    @simpleProperty
    def setScript(self):
        """String"""
        return PropertyType.STRING, "script"

    @multiProperty
    def setFrame(self):
        """Tuple of x,y,w,h"""
        return PropertyType.FRAME, "frame", ("x", "y", "w", "h")

    @multiProperty
    def setColor(self):
        """r, g, b, a"""
        return PropertyType.COLOR, "color", ("r", "g", "b", "a")

    @booleanProperty
    def setLocked(self):
        """Boolean"""
        return PropertyType.BOOLEAN, "locked"

    @booleanProperty
    def setBackground(self):
        """Boolean"""
        return PropertyType.BOOLEAN, "background"

    @booleanProperty
    def setVisible(self):
        """Boolean"""
        return PropertyType.BOOLEAN, "visible"

    @booleanProperty
    def setInteractive(self):
        """Boolean"""
        return PropertyType.BOOLEAN, "interactive"

    @booleanProperty
    def setOutline(self):
        """Boolean"""
        return PropertyType.BOOLEAN, "outline"

    def show(self):
        showElement(self.node)

    def showProperty(self, name: str):
        try:
            showElement(findKey(self.properties, name))
        except TypeError:
            raise ValueError(f"{name} doesn't exist")

    def showValue(self, name: str):
        try:
            showElement(findKey(self.values, name))
        except TypeError:
            raise ValueError(f"{name} doesn't exist")


"""

GENERAL FUNCTIONS

"""


def findKey(elements: ElementXML, key: str) -> ElementXML | Any:
    """Iterate through element with children and return child whose key matches"""
    return elements.find(f"*[key='{key}']")
    # for e in elements:
    #     if re.fullmatch(e.find("key").text, key):
    #         return e
    # return None


def showElement(e: ElementXML | None):
    """Generic print string function, UTF-8, indented 2 spaces"""
    if e is not None:
        ET.indent(e, "  ")
        print(ET.tostring(e).decode("utf-8"))


def createTemplate(frame: tuple = None) -> ElementXML:
    """Generates a root xml Element and adds the base GROUP node to it."""
    root = ET.Element("lexml", attrib={"version": "3"})
    node = ET.SubElement(
        root,
        ControlElements.NODE.value,
        attrib={"ID": str(uuid.uuid4()), "type": ControlType.GROUP.value},
    )
    if frame is not None:
        ElementTOSC(node).setFrame(frame)
    return root


def createGroup() -> ElementXML:
    """Simple create Node type GROUP"""
    return ET.Element(
        ControlElements.NODE.value,
        attrib={"ID": str(uuid.uuid4()), "type": ControlType.GROUP.value},
    )


def load(inputPath: str) -> ElementXML:
    """Reads a .tosc file and returns the XML root Element"""
    with open(inputPath, "rb") as file:
        return ET.fromstring(zlib.decompress(file.read()))


def write(root: ElementXML, outputPath: str) -> bool:
    """Encodes a root Element to .tosc"""
    with open(outputPath, "wb") as file:
        treeFile = ET.tostring(root, encoding="UTF-8", method="xml")
        file.write(zlib.compress(treeFile))
    return True


def getTextValueFromKey(properties: ElementXML | Any, key: str) -> str | Any:
    """Find the value.text from a known key"""
    if prop:= properties.find(f"./property/[key='{key}']"):
        value = prop.find("value")
        return value.text if value is not None else ""
    return None

def asCtrl(xml:ElementXML) -> Control:
    pass

def asXml(source: Control) -> ElementXML:
    return ControlConverter.build(source)

def asRoot(source: Control) -> ElementXML:
    root = ET.Element("lexml", attrib={"version": "3"})
    root.append(asXml(source))
    return root

def asEtosc(source: ElementXML | Control) -> ElementTOSC:
    if isinstance(source, ElementXML):
        return ElementTOSC(source)
    else:
        return ElementTOSC(ControlConverter.build(source))


"""
COPY AND MOVE
"""


def copyProperties(source: ElementTOSC, target: ElementTOSC, *args: str):
    """Args can be any number of property keys"""
    if args is None:
        [target.properties.append(deepcopy(e)) for e in source.properties]
        return True
    for arg in args:
        if elements := source.properties.findall(f"*[key='{arg}']"):
            [target.properties.append(deepcopy(e)) for e in elements]
        else:
            raise ValueError(f"Failed to find all elements with {args}")
    return True


def moveProperties(source: ElementTOSC, target: ElementTOSC, *args):
    elements = []
    if args is None:
        elements = source.properties
    for arg in args:
        if e := source.properties.findall(f"*[key='{arg}']"):
            elements += e
        else:
            raise ValueError(f"Failed to find all elements with {args}")

    [target.properties.append(deepcopy(e)) for e in elements]
    [source.properties.remove(e) for e in elements]
    return True


def copyValues(source: ElementTOSC, target: ElementTOSC, *args: str):
    """Args can be any number of value keys"""
    if args is None:
        [target.values.append(deepcopy(e)) for e in source.values]
        return True
    for arg in args:
        if elements := source.values.findall(f"*[key='{arg}']"):
            [target.values.append(deepcopy(e)) for e in elements]
        else:
            raise ValueError(f"Failed to find all elements with {args}")
    return True


def moveValues(source: ElementTOSC, target: ElementTOSC, *args: str):
    elements = []
    if args is None:
        elements = source.values
    for arg in args:
        if e := source.values.findall(f"*[key='{arg}']"):
            elements += e
        else:
            raise ValueError(f"Failed to find all elements with {args}")

    [target.values.append(deepcopy(e)) for e in elements]
    [source.values.remove(e) for e in elements]
    return True


def copyMessages(source: ElementTOSC, target: ElementTOSC, *args: elementType):
    """Args can be ControlElements.OSC, MIDI, LOCAL, GAMEPAD"""
    if args is None:
        [target.messages.append(deepcopy(e)) for e in source.messages]
        return True
    for arg in args:
        if elements := source.messages.findall(f"./{arg.value}"):
            [target.messages.append(deepcopy(e)) for e in elements]
        else:
            raise ValueError(f"Failed to find all elements with {arg.value}")
    return True


def moveMessages(source: ElementTOSC, target: ElementTOSC, *args: elementType):
    elements = []
    if args is None:
        elements = source.messages
    for arg in args:
        if e := source.messages.findall(f"./{arg.value}"):
            elements += e
        else:
            raise ValueError(f"Failed to find all elements with {arg.value}")

    [target.messages.append(deepcopy(e)) for e in elements]
    [source.messages.remove(e) for e in elements]
    return True


def copyChildren(source: ElementTOSC, target: ElementTOSC, *args: elementType):
    """Args can be ControlType.BOX, BUTTON, etc."""
    if args is None:
        [target.children.append(deepcopy(e)) for e in source.children]
        return True
    for arg in args:
        if elements := source.children.findall(
            f"./{ControlElements.NODE.value}[@type='{arg.value}']"
        ):
            [target.children.append(deepcopy(e)) for e in elements]
        else:
            raise ValueError(f"Failed to find all elements with {arg.value}")
    return True


def moveChildren(source: ElementTOSC, target: ElementTOSC, *args: elementType):
    elements = []
    if args is None:
        elements = source.children
    for arg in args:
        if e := source.children.findall(
            f"./{ControlElements.NODE}[@type='{arg.value}']"
        ):
            elements += e
        else:
            raise ValueError(f"Failed to find all elements with {arg.value}")

    [target.children.append(deepcopy(e)) for e in elements]
    [source.children.remove(e) for e in elements]
    return True


"""

GENERAL PARSERS

"""


def pullValueFromKey(inputFile: str, key: str, value: str,
                     targetKey: str) -> str | None:
    """If you know the name of an element but don't know its other properties.
    This function uses a .tosc file and gets its root.
    For passing an element see pullValueFromKey2

    Args:
        inputFile (str): File to parse.
        key (str): Known key.
        value (str): Known value.
        targetKey (str): Known key of unknown value.

    Returns:
        str: Value
    """
    parser = ET.XMLPullParser()
    with open(inputFile, "rb") as file:
        parser.feed(zlib.decompress(file.read()))
        for _, e in parser.read_events():  # event, element
            if e.find("properties") is None:
                continue
            if re.fullmatch(getTextValueFromKey(
                    e.find("properties"), key), value):
                parser.close()
                return getTextValueFromKey(e.find("properties"), targetKey)

    parser.close()
    return None


def pullValueFromKey2(root: ElementXML, key: str,
                      value: str, targetKey: str) -> str | None:
    """If you know the name of an element but don't know its other properties.
    This parses an Element and has to convert it to string so its slower.

    Args:
        root (ET.Element): Parses the whole element, so you can feed the root.
        key (str): Known key.
        value (str): Known value.
        targetKey (str): Known key of unknown value.

    Returns:
        str: Value
    """
    parser = ET.XMLPullParser()
    parser.feed(ET.tostring(root, encoding="UTF-8"))
    for _, e in parser.read_events():  # event, element
        if e.find("properties") is None:
            continue
        if re.fullmatch(getTextValueFromKey(e.find("properties"), key), value):
            parser.close()
            return getTextValueFromKey(e.find("properties"), targetKey)
    return None

def pullIdfromName(root: ElementXML, name: str) -> str:
    """
    """
    parser = ET.XMLPullParser()
    parser.feed(ET.tostring(root, encoding="UTF-8"))
    for _, e in parser.read_events():  # event, element
        if e.find("properties") is None:
            continue
        # xpath ftw
        if e.find(f"./properties/property/[value='{name}']") is not None:            
            parser.close()
            return e.attrib["ID"]
    
    raise ValueError(f"{name}'s ID wasn't found.")



class PropertyParser:
    """Find all defined properties in the Node"""

    def __init__(self, *args):
        self.targetList = []
        self.args = [*args]
        self.targetFound = None
        self.multiLine = ""
        self.node = False
        self.property = False
        self.key = False
        self.value = False
        self.index = -1

    def start(self, tag, attrib):
        if tag == ControlElements.NODE.value:
            self.index += 1
            self.node = True
            self.targetList.append({arg: "" for arg in [*self.args]})
        elif self.node and tag == ControlElements.PROPERTY.value:
            self.property = True
        elif self.property and tag == "key":
            self.key = True
        elif self.property and tag == "value":
            self.value = True

    def end(self, tag):
        if tag == ControlElements.NODE.value:
            self.node = False
        elif self.node and tag == ControlElements.PROPERTY.value:
            self.property = False
        elif self.property and tag == "key":
            self.key = False
        elif self.property and tag == "value":
            self.value = False

        if self.targetFound and tag == "value":
            self.targetList[self.index][self.targetFound] = self.multiLine
            self.multiLine = ""
            self.targetFound = None

    def data(self, data):
        if (
            self.node
            and self.property
            and self.key
            # and data in self.targetList.keys()
            and data in self.args
        ):
            self.targetFound = data
        if self.node and self.property and self.value and self.targetFound:
            self.multiLine = f"{self.multiLine}{data}"

    def close(self):
        return self.targetList


def parseProperties(node: ElementXML, *args) -> list:
    """
    Specify all properties you want to find and this will parse
    the entire Node and its children and return a list of key value pairs.

    For example:

    [{"name":"control1", "script":"scriptContent1"},
    {"name":"control2", "script":""},
    {"name":"control3", "script":"scriptContent3"}]

    Args:
        node (ET.Element): Node element to parse.

    Returns:
        List[dict]: [{arg: "" for arg in [args]}]
    """
    target = PropertyParser(*args)
    line = ET.tostring(node, encoding="UTF-8")
    parser = ET.XMLParser(target=target)
    parser.feed(line)
    return parser.close()


"""

Testing stuff

"""


def testFromString(data):
    return ET.fromstring(data)



""" Enumerations for constructing XML Elements"""

from copy import deepcopy
from enum import Enum, unique
from typing import Dict, Literal, TypeAlias, TypeGuard, TypedDict

"""

FIRST SECTION:

Define Enumerations and Literal types from Hexler's Touch OSC design.

SECOND SECTION:

Define struct like classes that carry the data to create xml Elements.
Plus Factory classes to create them.

"""


@unique
class ControlElements(Enum):
    """Valid xml tags for a Control"""

    PROPERTIES = "properties"  #: <properties>
    VALUES = "values"  #: <values>
    MESSAGES = "messages"  #: <messages>
    CHILDREN = "children"  #: <children>
    PROPERTY = "property"  #: <property type=>
    VALUE = "value"  #: <value>
    OSC = "osc"  #: <osc>
    MIDI = "midi"  #: <midi>
    LOCAL = "local"  #: <local>
    GAMEPAD = "gamepad"  #: <gamepad>
    NODE = "node"  #: <node type =>


@unique
class ControlType(Enum):
    """Valid xml attrib = {"type":ControlType} for <node>"""

    BOX = "BOX"  #: <node type = "BOX">
    BUTTON = "BUTTON"  #: <node type = "BUTTON">
    LABEL = "LABEL"  #: <node type = "LABEL">
    TEXT = "TEXT"  #: <node type = "TEXT">
    FADER = "FADER"  #: <node type = "FADER">
    XY = "XY"  #: <node type = "XY">
    RADIAL = "RADIAL"  #: <node type = "RADIAL">
    ENCODER = "ENCODER"  #: <node type = "ENCODER">
    RADAR = "RADAR"  #: <node type = "RADAR">
    RADIO = "RADIO"  #: <node type = "RADIO">
    GROUP = "GROUP"  #: <node type = "GROUP">
    PAGER = "PAGER"  #: <node type = "PAGER">
    GRID = "GRID"  #: <node type = "GRID">


@unique
class PropertyType(Enum):
    """Valid xml attrib = {"type":PropertyType} for <property>"""

    STRING = "s"  #: <property type="s">
    BOOLEAN = "b"  #: <property type="b">
    INTEGER = "i"  #: <property type="i">
    FLOAT = "f"  #: <property type="f">
    FRAME = "r"  #: <property type="r">
    COLOR = "c"  #: <property type="c">


elementType = Literal[
    ControlElements.CHILDREN,
    ControlElements.GAMEPAD,
    ControlElements.LOCAL,
    ControlElements.MESSAGES,
    ControlElements.MIDI,
    ControlElements.NODE,
    ControlElements.OSC,
    ControlElements.PROPERTIES,
    ControlElements.PROPERTY,
    ControlElements.VALUES,
    ControlElements.VALUE,
]


controlType = Literal[
    ControlType.BOX,
    ControlType.BUTTON,
    ControlType.ENCODER,
    ControlType.FADER,
    ControlType.GRID,
    ControlType.GROUP,
    ControlType.LABEL,
    ControlType.PAGER,
    ControlType.RADAR,
    ControlType.RADIAL,
    ControlType.RADIO,
    ControlType.TEXT,
    ControlType.XY,
]

propertyType = Literal[
    PropertyType.BOOLEAN,
    PropertyType.COLOR,
    PropertyType.FLOAT,
    PropertyType.FRAME,
    PropertyType.INTEGER,
    PropertyType.STRING,
]


class Frame(TypedDict):
    """Typed dictionary of 4 floats"""

    x: int
    y: int
    w: int
    h: int


class Color(TypedDict):
    """Typed dictionary of 4 floats"""

    r: float
    g: float
    b: float
    a: float


"""

SECOND SECTION:

"""


class Property:
    """Struct like object to carry the property values"""

    __slots__ = ("type", "key", "value", "params")

    def __init__(
        self, type: str, key: str, value: str | None = None, params: dict[str,str] = None
    ):
        self.type: str = type
        self.key: str = key
        self.value: str | None = value if value is not None else value
        self.params: dict[str, str] | dict = params if params is not None else {}


    def __repr__(self):
        return f"""
Property:
    type:{self.type}
    key:{self.key}
    value:{self.value}
    params:{self.params}"""

class Value:
    """Default Elements for <value>.

    Args:
        key (str, optional): "x" or "touch". Defaults to "touch".
        locked (str, optional): boolean. Defaults to "0".
        lockedDefaultCurrent (str, optional): boolean. Defaults to "0".
        default (str, optional): float or boolean. Defaults to "false".
        defaultPull (str, optional): 0 to 100. Defaults to "0".
    """

    __slots__ = ("key", "locked", "lockedDefaultCurrent", "default", "defaultPull")

    def __init__(
        self,
        key: str = "touch",
        locked: str = "0",
        lockedDefaultCurrent: str = "0",
        default: str = "false",
        defaultPull: str = "0",
    ):
        self.key: str = key
        self.locked: str = locked
        self.lockedDefaultCurrent: str = lockedDefaultCurrent
        self.default: str = default
        self.defaultPull: str = defaultPull


class Partial:
    """Default Elements for <partial>

    Args:
        type (str, optional):
        "CONSTANT", "INDEX", "VALUE", "PROPERTY". Defaults to "CONSTANT".
        conversion (str, optional):
        "BOOLEAN", "INTEGER", "FLOAT", "STRING". Defaults to "STRING".
        value (str, optional): Depends on the context. Defaults to "/".
        scaleMin (str, optional): If "VALUE", set range. Defaults to "0".
        scaleMax (str, optional): If "VALUE", set range. Defaults to "1".
    """

    __slots__ = ("type", "conversion", "value", "scaleMin", "scaleMax")

    def __init__(
        self,
        type: str = "CONSTANT",
        conversion: str = "STRING",
        value: str = "/",
        scaleMin: str = "0",
        scaleMax: str = "1",
    ):
        self.type = type
        self.conversion = conversion
        self.value = value
        self.scaleMin = scaleMin
        self.scaleMax = scaleMax


class Trigger:
    """Default Elements for <trigger>

    Args:
        var (str, optional): "x" or "touch". Defaults to "x".
        con (str, optional): "ANY", "RISE" or "FALL". Defaults to "ANY".
    """

    __slots__ = ("var", "condition")

    def __init__(self, var: str = "x", condition: str = "ANY"):
        self.var = var
        self.condition = condition


class MidiMessage:
    """_summary_

    Args:
        type (str, optional): CONTROLCHANGE, NOTE_ON/OFF, etc.
        Defaults to "CONTROLCHANGE".
        channel (str, optional): Midi Channel. Defaults to "1".
        data1 (str, optional): Depends on type. Defaults to "0".
        data2 (str, optional): Depends on type. Defaults to "0".
    """

    __slots__ = ("type", "channel", "data1", "data2")

    def __init__(
        self,
        type: str = "CONTROLCHANGE",
        channel: str = "1",
        data1: str = "0",
        data2: str = "0",
    ):
        self.type = type
        self.channel = channel
        self.data1 = data1
        self.data2 = data2


class MidiValue:
    """The Value of the control to send as Midi, generally 0-127 scale.

    Args:
        type (str, optional): CONSTANT, INDEX, VALUE, PROPERTY.
        Defaults to "VALUE".
        key (str, optional): Value or Property of the control. Defaults to "x".
        scaleMin (str, optional): Scale of the control. Defaults to "0".
        scaleMax (str, optional): Scale of the control. Defaults to "127".
    """

    __slots__ = ("type", "key", "scaleMin", "scaleMax")

    def __init__(
        self,
        type: str = "VALUE",
        key: str = "x",
        scaleMin: str = "0",
        scaleMax: str = "127",
    ):

        self.type = type
        self.key = key
        self.scaleMin = scaleMin
        self.scaleMax = scaleMax


class OSC:
    """Default Elements and Sub Elements for <osc>

    Args:
        enabled (str, optional): Boolean. Defaults to "1".
        send (str, optional): Boolean. Defaults to "1".
        receive (str, optional): Boolean. Defaults to "1".
        feedback (str, optional): Boolean. Defaults to "0".
        connections (str, optional): Binary.
        Defaults to "00001" (channel 1, "00011" means 1 and 2).
        triggers (List[Trigger], optional): [Trigger].
        Defaults to [Trigger()].
        path (List[Partial], optional): [Partial].
        Defaults to [Partial(), Partial(typ="PROPERTY", val="name")].
        arguments (List[Partial], optional): [Partial].
        Defaults to [Partial(typ="VALUE", con="FLOAT", val="x")].
    """

    __slots__ = (
        "enabled",
        "send",
        "receive",
        "feedback",
        "connections",
        "triggers",
        "path",
        "arguments",
    )

    def __init__(
        self,
        enabled: str = "1",
        send: str = "1",
        receive: str = "1",
        feedback: str = "0",
        connections: str = "00001",
        triggers: list[Trigger] = [Trigger()],
        path: list[Partial] = [Partial(), Partial(type="PROPERTY", value="name")],
        arguments: list[Partial] = [
            Partial(type="VALUE", conversion="FLOAT", value="x")
        ],
    ):
        self.enabled = enabled
        self.send = send
        self.receive = receive
        self.feedback = feedback
        self.connections = connections
        self.triggers = triggers
        self.path = path
        self.arguments = arguments


class MIDI:
    __slots__ = (
        "enabled",
        "send",
        "receive",
        "feedback",
        "connections",
        "triggers",
        "message",
        "values",
    )

    def __init__(
        self,
        enabled: str = "1",
        send: str = "1",
        receive: str = "1",
        feedback: str = "0",
        connections: str = "00001",
        triggers: list[Trigger] = [Trigger()],
        message: MidiMessage = MidiMessage(),
        values: list[MidiValue] = [
            MidiValue(),
            MidiValue("INDEX", "", "0", "1"),
            MidiValue("VALUE", "x", "0", "127"),
        ],
    ):
        self.enabled = enabled
        self.send = send
        self.receive = receive
        self.feedback = feedback
        self.connections = connections
        self.triggers = triggers
        self.message = message
        self.values = values


class LOCAL:
    """Default elements for <midi>
    Args:
        enabled: bool
        triggers : Trigger x or touch.
        type : BOOL, INT, FLOAT, STRING. The Type of Trigger.x
        conversion : BOOL, INT, FLOAT, STRING.
        value : The value sent to the other local Control.
        scaleMin : 0
        scaleMax : 1
        dstType : BOOL, INT, FLOAT, STRING of the target.
        dstVar : The value you want to change in the target.
        dstID : The node {ID} of the target.
    """

    __slots__ = (
        "enabled",
        "triggers",
        "type",
        "conversion",
        "value",
        "scaleMin",
        "scaleMax",
        "dstType",
        "dstVar",
        "dstID",
    )

    def __init__(
        self,
        enabled: str = "1",
        triggers: list[Trigger] = [Trigger()],
        type: str = "VALUE",
        conversion: str = "FLOAT",
        value: str = "x",
        scaleMin: str = "0",
        scaleMax: str = "1",
        dstType: str = "",
        dstVar: str = "",
        dstID: str = "",
    ):
        self.enabled = enabled
        self.triggers = triggers
        self.type = type
        self.conversion = conversion
        self.value = value
        self.scaleMin = scaleMin
        self.scaleMax = scaleMax
        self.dstType = dstType
        self.dstVar = dstVar
        self.dstID = dstID


class PropertyFactory:
    @classmethod
    def build(
        cls, 
        key:str, 
        value: int | bool | float | str | tuple[int,...] | tuple[float,...]
    ) -> Property:
        """_summary_

        Args:
            key (_type_): _description_
            value (int | bool | float | str | tuple[int] | tuple[float]):
            Whatever value with one of those python types.
            Tuples of ints are considered Frames.
            Tuples of floats are considered Colors.

        Raises:
            ValueError:
            If value type is not compatible with TouchOSC's equivalent.

        Returns:
            Property: _description_
        """
        if isinstance(value, str):
            return cls.buildString(key, value)
        elif isinstance(value, bool):
            return cls.buildBoolean(key, value)
        elif isinstance(value, int):
            return cls.buildInteger(key, value)
        elif isinstance(value, float):
            return cls.buildFloat(key, value)
        elif isinstance(value, tuple) and cls.isTupleInts(value):
            return cls.buildFrame(key, value)
        elif isinstance(value, tuple) and all(isinstance(x, float) for x in value):
            return cls.buildColor(key, value)
        else:
            raise ValueError(f"""
{key}-{value} type is not a valid PropertyType.
Make sure all types in the tuples match.""")

    @classmethod
    def isTupleInts(cls, val:tuple[object,...]) -> TypeGuard[tuple[int,...]]:
        return all(isinstance(x, int) for x in val)

    @classmethod
    def buildBoolean(cls, key: str, value: bool) -> Property:
        return Property(PropertyType.BOOLEAN.value, key, repr(int(value)))

    @classmethod
    def buildString(cls, key: str, value: str) -> Property:
        return Property(PropertyType.STRING.value, key, value)

    @classmethod
    def buildInteger(cls, key: str, value: int) -> Property:
        return Property(PropertyType.INTEGER.value, key, repr(value))

    @classmethod
    def buildFloat(cls, key: str, value: float) -> Property:
        return Property(PropertyType.FLOAT.value, key, repr(value))

    @classmethod
    def buildColor(cls, key: str, value: tuple[float,...]) -> Property:
        return Property(
            PropertyType.COLOR.value,
            key,
            "",
            {
                "r": repr(value[0]),
                "g": repr(value[1]),
                "b": repr(value[2]),
                "a": repr(value[3]),
            },
        )

    @classmethod
    def buildFrame(cls, key: str, value: tuple[int,...]) -> Property:
        return Property(
            PropertyType.FRAME.value,
            key,
            "",
            {
                "x": repr(value[0]),
                "y": repr(value[1]),
                "w": repr(value[2]),
                "h": repr(value[3]),
            },
        )

    @classmethod
    def name(cls, value: str) -> Property:
        return cls.buildString("name", value)

    @classmethod
    def tag(cls, value: str) -> Property:
        return cls.buildString("tag", value)

    @classmethod
    def script(cls, value: str) -> Property:
        return cls.buildString("script", value)

    @classmethod
    def frame(cls, params: tuple[int,...]) -> Property:
        return cls.buildFrame("frame", params)

    @classmethod
    def color(cls, params: tuple[float,...]) -> Property:
        return cls.buildColor("color", params)

    @classmethod
    def locked(cls, value: bool) -> Property:
        return cls.buildBoolean("locked", value)

    @classmethod
    def visible(cls, value: bool) -> Property:
        return cls.buildBoolean("visible", value)

    @classmethod
    def interactive(cls, value: bool) -> Property:
        return cls.buildBoolean("interactive", value)

    @classmethod
    def background(cls, value: bool) -> Property:
        return cls.buildBoolean("background", value)

    @classmethod
    def outline(cls, value: bool) -> Property:
        return cls.buildBoolean("outline", value)

    @classmethod
    def outlineStyle(cls, value: int) -> Property:
        return cls.buildInteger("outline", value)

    @classmethod
    def textColor(cls, params: tuple[float,...]) -> Property:
        return cls.buildColor("textColor", params)

    @classmethod
    def textSize(cls, value: int) -> Property:
        return cls.buildInteger("textSize", value)


class ValueFactory:
    @classmethod
    def build(
        cls, 
        key:str, 
        value: int | bool | float | str | tuple[int,...] | tuple[float,...]
    ):
        return


import setuptools # type: ignore

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="tosclib",
    version="0.3.5",
    author="Alberto Valdez",
    author_email="avq5ac1@gmail.com",
    description="Generate and edit Touch OSC .tosc files",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/AlbertoV5/tosclib",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    package_data={"tosclib": ["py.typed"]},
    python_requires=">=3.10",
    install_requires=[
        "lxml>=4.9.0",
        "numpy>=1.22.3"
    ]
)


"""
Hexler's Enumerations
"""

from copy import deepcopy
from dataclasses import dataclass, field
from types import UnionType
from typing import ClassVar, Final, Protocol, TypeAlias, TypeGuard
import uuid
from .elements import (
    MidiMessage,
    Property,
    Value,
    OSC,
    MIDI,
    LOCAL,
    PropertyFactory,
    controlType,
    ControlType,
    ControlElements,
)

import xml.etree.ElementTree as ET
# from lxml import etree as ET

Properties: TypeAlias = list[Property]
Values: TypeAlias = list[Value]
Message: TypeAlias = OSC | MIDI | LOCAL
Messages: TypeAlias = list[OSC | MIDI | LOCAL]


@dataclass
class _ControlProperties:
    """Common properties across all Control Types
    https://hexler.net/touchosc/manual/script-properties-and-values"""

    name: Final[str] = " "
    """Any string"""
    tag: Final[str] = "tag"
    """Any string"""
    script: Final[str] = " "
    """Any string"""
    frame: Final[tuple] = field(default_factory=lambda: (0, 0, 100, 100))
    """x,y,w,h float list"""
    color: Final[tuple] = field(default_factory=lambda: (0.25, 0.25, 0.25, 1.0))
    """r,g,b,a float list"""
    locked: Final[bool] = False
    visible: Final[bool] = True
    interactive: Final[bool] = True
    background: Final[bool] = True
    outline: Final[bool] = True
    outlineStyle: int = 1
    """0,1,2, = Full, Corner, Edges"""
    grabFocus: bool = True
    """Depends on the control, groups are false"""
    pointerPriority: Final[int] = 0
    """0,1 = Oldest, Newest"""
    cornerRadius: Final[float] = 0.0
    """An integer number value ranging from 0 to 10"""
    orientation: int = 0
    """0,1,2,3 = North, East, South, West"""

    def build(self, *args) -> Properties:
        """Build all Property objects of this class.
        Returns:
            list[Property] from this class' attributes.
        """
        if len(args) == 0:
            args = tuple(key for key in vars(self))

        return [PropertyFactory.build(arg, getattr(self, arg)) for arg in args]


@dataclass
class _BoxProperties:
    shape: int = 0
    """0,1,2,3,4,5 Rectangle, Circle, Triangle, Diamond, Pentagon, Hexagon"""


@dataclass
class _GroupProperties:
    outlineStyle: int = 0
    """0,1,2, = Full, Corner, Edges"""
    grabFocus: bool = False
    """Depends on the control, groups are false"""


@dataclass
class _GridProperties:
    grid: Final[bool] = True
    gridSteps: Final[int] = 10
    """Size of grid"""


@dataclass
class _ResponseProperties:
    response: Final[int] = 0
    """0,1 = Absolute, Relative"""
    responseFactor: Final[int] = 100
    """An integer value ranging from 1 to 100."""


@dataclass
class _CursorProperties:
    cursor: Final[bool] = True
    cursorDisplay: Final[int] = 0
    """Cursor display 0, 1, 2 = always, active, inactive"""


@dataclass
class _LineProperties:
    lines: Final[bool] = True
    linesDisplay: Final[int] = 0
    """Cursor display 0, 1, 2 = always, active, inactive"""


@dataclass
class _XyProperties:
    lockX: Final[bool] = False
    lockY: Final[bool] = False
    gridX: Final[bool] = True
    gridY: Final[bool] = True
    gridStepsX: Final[int] = 10
    gridStepsY: Final[int] = 10


@dataclass
class _TextProperties:
    font: int = 0
    """0, 1 = default, monospaced"""
    textSize: Final[int] = 14
    """Any int"""
    textColor: Final[tuple] = field(default_factory=lambda: (1, 1, 1, 1))
    """rgba dict from 0 to 1 as str"""
    textAlignH: Final[int] = 2
    """1,2,3 = left, center, right"""


@dataclass
class BoxProperties(_ControlProperties, _BoxProperties):
    orientation: int = 0
    """0,1,2,3 = North, East, South, West"""


@dataclass
class ButtonProperties(_ControlProperties, _BoxProperties):
    buttonType: Final[int] = 0
    """0,1,2 Momentary, Toggle_Release, Toggle_Press"""
    press: Final[bool] = True
    release: Final[bool] = True
    valuePosition: Final[bool] = False


@dataclass
class LabelProperties(_ControlProperties, _TextProperties):
    textLength: Final[int] = 0
    """0 is infinite length"""
    textClip: Final[bool] = True


@dataclass
class TextProperties(_ControlProperties, _TextProperties):
    pass


@dataclass
class FaderProperties(
    _ControlProperties, _ResponseProperties, _GridProperties, _CursorProperties
):
    bar: Final[bool] = True
    barDisplay: Final[int] = 0


@dataclass
class XyProperties(
    _ControlProperties,
    _ResponseProperties,
    _CursorProperties,
    _XyProperties,
):
    pass


@dataclass
class RadialProperties(
    _ControlProperties,
    _ResponseProperties,
    _GridProperties,
    _CursorProperties,
):
    outlineStyle: int = 0
    """0,1,2, = Full, Corner, Edges"""
    inverted: Final[bool] = False
    centered: Final[bool] = False


@dataclass
class EncoderProperties(_ControlProperties, _ResponseProperties, _GridProperties):
    outlineStyle: int = 0
    """0,1,2, = Full, Corner, Edges"""


@dataclass
class RadarProperties(
    _ControlProperties,
    _CursorProperties,
    _LineProperties,
    _XyProperties,
):
    pass


@dataclass
class RadioProperties(_ControlProperties):
    steps: Final[int] = 5
    """Amount of radio steps"""
    radioType: Final[int] = 0
    """0,1 = select, meter"""
    orientation: int = 0
    """0,1,2,3 = North, East, South, West"""


@dataclass
class GroupProperties(_ControlProperties, _GroupProperties):
    pass


@dataclass
class GridProperties(_ControlProperties):
    grabFocus: bool = False
    """Depends on the control, groups are false"""
    exclusive: bool = False
    gridNaming: Final[int] = 0
    """0,1,2 = Index, Column, Row"""
    gridOrder: Final[int] = 0
    """0,1 = Row, Column"""
    gridStart: Final[int] = 0
    """0,1,2,3 = Top left, Top right, Bottom Left, Bottom Right"""
    gridType: Final[int] = 4
    """0,1,2,3,4,5,6,7,8 See ControlType, can't hold groups"""
    gridX: Final[int] = 2
    """amount of elements on X"""
    gridY: Final[int] = 2
    """amount of elements on Y"""


@dataclass
class PagerProperties(_ControlProperties, _GroupProperties):
    """0,1,2, = Full, Corner, Edges"""

    tabLabels: Final[bool] = True
    tabbar: Final[bool] = True
    tabbarDoubleTap: Final[bool] = False
    tabbarSize: Final[int] = 40
    """int from 10 to 300"""
    textSizeOff: Final[int] = 14
    """font size any int"""
    textSizeOn: Final[int] = 14
    """font size any int"""


@dataclass
class PageProperties(_ControlProperties, _GroupProperties):
    tabColorOff: Final[tuple] = field(default_factory=lambda: (0.25, 0.25, 0.25, 1.0))
    tabColorOn: Final[tuple] = field(default_factory=lambda: (0, 0, 0, 0))
    tabLabel: Final[str] = "1"
    textColorOff: Final[tuple] = field(default_factory=lambda: (1, 1, 1, 1))
    textColorOn: Final[tuple] = field(default_factory=lambda: (1, 1, 1, 1))


@dataclass
class Page:
    """Not a main control"""

    controlT: ClassVar[controlType] = ControlType.GROUP
    id: str = field(default_factory=lambda:str(uuid.uuid4()))
    properties: Properties = field(default_factory=lambda:PageProperties().build())
    values: Values = field(default_factory=lambda: [])
    messages: Messages = field(default_factory=lambda: [])
    children: list["Control"] | None | list = None


@dataclass
class Box:
    controlT: ClassVar[controlType] = ControlType.BOX
    id: str = field(default_factory=lambda:str(uuid.uuid4()))
    properties: Properties = field(default_factory=lambda: BoxProperties().build())
    values: Values = field(default_factory=lambda: [])
    messages: Messages = field(default_factory=lambda: [])
    children: list["Control"] | None | list = None


@dataclass
class Button:
    controlT: ClassVar[controlType] = ControlType.BUTTON
    id: str = field(default_factory=lambda:str(uuid.uuid4()))
    properties: Properties = field(
        default_factory=lambda: ButtonProperties().build()
    )
    values: Values = field(default_factory=lambda: [])
    messages: Messages = field(default_factory=lambda: [])
    children: list["Control"] | None | list = None


@dataclass
class Label:
    controlT: ClassVar[controlType] = ControlType.LABEL
    id: str = field(default_factory=lambda:str(uuid.uuid4()))
    properties: Properties = field(
        default_factory=lambda: LabelProperties().build()
    )
    values: Values = field(default_factory=lambda: [])
    messages: Messages = field(default_factory=lambda: [])
    children: list["Control"] | None | list = None


@dataclass
class Text:
    controlT: ClassVar[controlType] = ControlType.TEXT
    id: str = field(default_factory=lambda:str(uuid.uuid4()))
    properties: Properties = field(
        default_factory=lambda: TextProperties().build()
    )
    values: Values = field(default_factory=lambda: [])
    messages: Messages = field(default_factory=lambda: [])
    children: list["Control"] | None | list = None


@dataclass
class Fader:
    controlT: ClassVar[controlType] = ControlType.FADER
    id: str = field(default_factory=lambda:str(uuid.uuid4()))
    properties: Properties = field(
        default_factory=lambda: FaderProperties().build()
    )
    values: Values = field(default_factory=lambda: [])
    messages: Messages = field(default_factory=lambda: [])
    children: list["Control"] | None | list = None


@dataclass
class Xy:
    controlT: ClassVar[controlType] = ControlType.XY
    id: str = field(default_factory=lambda:str(uuid.uuid4()))
    properties: Properties = field(default_factory=lambda: XyProperties().build())
    values: Values = field(default_factory=lambda: [])
    messages: Messages = field(default_factory=lambda: [])
    children: list["Control"] | None | list = None


@dataclass
class Radial:
    controlT: ClassVar[controlType] = ControlType.RADIAL
    id: str = field(default_factory=lambda:str(uuid.uuid4()))
    properties: Properties = field(
        default_factory=lambda: RadialProperties().build()
    )
    values: Values = field(default_factory=lambda: [])
    messages: Messages = field(default_factory=lambda: [])
    children: list["Control"] | None | list = None


@dataclass
class Encoder:
    controlT: ClassVar[controlType] = ControlType.ENCODER
    id: str = field(default_factory=lambda:str(uuid.uuid4()))
    properties: Properties = field(
        default_factory=lambda: EncoderProperties().build()
    )
    values: Values = field(default_factory=lambda: [])
    messages: Messages = field(default_factory=lambda: [])
    children: list["Control"] | None | list = None


@dataclass
class Radar:
    controlT: ClassVar[controlType] = ControlType.RADAR
    id: str = field(default_factory=lambda:str(uuid.uuid4()))
    properties: Properties = field(
        default_factory=lambda: RadarProperties().build()
    )
    values: Values = field(default_factory=lambda: [])
    messages: Messages = field(default_factory=lambda: [])
    children: list["Control"] | None | list = None


@dataclass
class Radio:
    controlT: ClassVar[controlType] = ControlType.RADIO
    id: str = field(default_factory=lambda:str(uuid.uuid4()))
    properties: Properties = field(
        default_factory=lambda: RadioProperties().build()
    )
    values: Values = field(default_factory=lambda: [])
    messages: Messages = field(default_factory=lambda: [])
    children: list["Control"] | None | list = None


@dataclass
class Group:
    controlT: ClassVar[controlType] = ControlType.GROUP
    id: str = field(default_factory=lambda:str(uuid.uuid4()))
    properties: Properties = field(
        default_factory=lambda: GroupProperties().build()
    )
    values: Values = field(default_factory=lambda: [])
    messages: Messages = field(default_factory=lambda: [])
    children: list["Control"] | None | list = None


@dataclass
class Grid:
    controlT: ClassVar[controlType] = ControlType.GRID
    id: str = field(default_factory=lambda:str(uuid.uuid4()))
    properties: Properties = field(
        default_factory=lambda: GridProperties().build()
    )
    values: Values = field(default_factory=lambda: [])
    messages: Messages = field(default_factory=lambda: [])
    children: list["Control"] | None | list = None


@dataclass
class Pager:
    controlT: ClassVar[controlType] = ControlType.PAGER
    id: str = field(default_factory=lambda:str(uuid.uuid4()))
    properties: Properties = field(
        default_factory=lambda: PagerProperties().build()
    )
    values: Values = field(default_factory=lambda: [])
    messages: Messages = field(default_factory=lambda: [])
    children: list["Control"] | None | list = field(
        default_factory=lambda: [Group(), Group(), Group()]
    )


class Control(Protocol):
    """Protocol type of Control

    Attributes:
        controlT: Control Type
        properties: List of Property
        values: List of Value
        messages: List of Message
    """
    controlT: ClassVar[controlType]
    id: str
    properties: Properties
    values: Values
    messages: Messages
    children: list["Control"] | None | list

# Control:TypeAlias = (
#     Label | Text | Box | Button | Group | Encoder | Page |
#     Grid | Fader | Pager | Radial | Radio | Radar | Xy)


class ControlProperties(Protocol):
    """Data structure to store control properties and defaults
    for reference."""
    name: str
    tag: str
    script: str
    frame: tuple
    color: tuple

    def build(self) -> Properties:
        ...



class ControlFactory:
    @classmethod
    def build(cls, 
    controlT: controlType, 
    id: str = None,
    properties: Properties = None,
    values: Values = None,
    messages: Messages = None) -> Control:
        id = str(uuid.uuid4()) if id is None else id
        properties = [] if properties is None else properties
        values = [] if values is None else values
        messages = [] if messages is None else messages
        match controlT:
            case ControlType.BOX:
                return Box(id, properties, values, messages)
            case ControlType.BUTTON:
                return Button(id, properties, values, messages)
            case ControlType.ENCODER:
                return Encoder(id, properties, values, messages)
            case ControlType.FADER:
                return Fader(id, properties, values, messages)
            case ControlType.GRID:
                return Grid(id, properties, values, messages)
            case ControlType.GROUP:
                return Group(id, properties, values, messages)
            case ControlType.LABEL:
                return Label(id, properties, values, messages)
            case ControlType.PAGER:
                return Pager(id, properties, values, messages)
            case ControlType.RADAR:
                return Radar(id, properties, values, messages)
            case ControlType.RADIAL:
                return Radial(id, properties, values, messages)
            case ControlType.RADIO:
                return Radio(id, properties, values, messages)
            case ControlType.TEXT:
                return Text(id, properties, values, messages)
            case ControlType.XY:
                return Xy(id, properties, values, messages)
            case _:
                raise ValueError(f"{controlT} not found.")


class ControlConverter:
    """Convert from Control to XML"""
    @classmethod
    def build(cls, control: Control) -> ET.Element:
        """Generate the XML Element and its SubElements

        Args:
            control (Control): Control

        Returns:
            ET.Element: XML
        """
        node = ET.Element(
            ControlElements.NODE.value,
            attrib={"ID": control.id, "type": control.controlT.value},
        )
        properties = ET.SubElement(node, ControlElements.PROPERTIES.value)
        values = ET.SubElement(node, ControlElements.VALUES.value)
        messages = ET.SubElement(node, ControlElements.MESSAGES.value)
        
        XmlFactory.buildProperties(properties, control.properties)
        XmlFactory.buildValues(values, control.values)
        XmlFactory.buildMessages(messages, control.messages)

        if control.children is not None:
            children = ET.SubElement(node, ControlElements.CHILDREN.value)
            for child in control.children:
                children.append(cls.build(child))
        
        return node


class XmlFactory:
    """Generate specific XML structures"""
    @classmethod
    def buildProperties(cls, e: ET.Element, props: Properties) -> bool:
        """Create many <property> XML

        Args:
            props (Properties): List of Properties
            e (ET.Element): Element to append the properties to

        Returns:
            bool: bool
        """
        for prop in props:
            property = ET.SubElement(
                e, ControlElements.PROPERTY.value, attrib={"type": prop.type}
            )
            ET.SubElement(property, "key").text = prop.key
            value = ET.SubElement(property, "value")
            value.text = prop.value
            for k in prop.params:
                ET.SubElement(value, k).text = prop.params[k]
        return True

    @classmethod
    def modifyProperty(cls, e: ET.Element, value:str, params:dict[str,str]) -> bool:
        """Modify an existing property XML

        Args:
            value (str): New value
            params (dict): New params
            p (ET.Element): <property>

        Returns:
            bool: bool
        """
        if (v:= e.find("value")) is not None:
            v.text = value
            for k in params:
                if (p := v.find(k)) is not None:
                    p.text = params[k]
            return True
        return False

    @classmethod
    def buildValues(cls, e: ET.Element, vals: Values) -> bool:
        for val in vals:
            value = ET.SubElement(e, ControlElements.VALUE.value)
            for k in val.__slots__:
                ET.SubElement(value, k).text = getattr(val, k)
        return True

    @classmethod
    def modifyValue(cls, e: ET.Element, val: Value) -> bool:
        for k in val.__slots__:
            if (v:=e.find(k)) is not None:
                v.text = getattr(val, k)
        return True

    @classmethod
    def buildMessages(cls, e: ET.Element, msgs: Messages) -> bool:
        for message in msgs:
            msg = ET.SubElement(e, message.__class__.__name__.lower())
            for k in message.__slots__:
                element = ET.SubElement(msg, k)
                v = getattr(message, k)
                if isinstance(v, list):
                    for pt in v:
                        e = ET.SubElement(element, type(pt).__name__.lower())
                        for x in pt.__slots__:
                            ET.SubElement(e, x).text = getattr(pt, x)
                elif isinstance(v, MidiMessage):
                    for x in pt.__slots__:
                        ET.SubElement(element, x).text = getattr(pt, x)
                else:
                    element.text = v
        return True

    @classmethod
    def buildNode(cls, e: ET.Element, controlT: controlType, ) -> ET.Element:
        return ET.SubElement(
            e,
            ControlElements.NODE.value,
            attrib={"ID": str(uuid.uuid4()), "type": controlT.value},
        )
